module Ascii85
  def self.encode( str )
    ires = 3 - ( str.length + 3 ) % 4
    base = (0...5).collect{ |i| 85 ** i }.reverse!
    ( str + "\x00\x00\x01"[-ires,ires] ).unpack( "N*" ).collect!{ |i|
      i == 0 ? [ 0x7a ] : base.collect{ |b| j, i = i.divmod( b ); j + 33 }
    }.flatten!.pack( "C*" )[0..(-1-ires)]
  end

  def self.decode( str )
    buf = str.gsub( "z", "!"*5 ).delete( "^!-u" )
    ires = 4 - ( buf.length + 4 ) % 5
    ( buf << "u" * ires ).unpack( "C*" ).each_slice(5).collect{ |a|
      a.inject( 0 ){ |i,c| i * 85 + ( c - 33 ) }
    }.pack( "N*" )[0..(-1-ires)]
  end
end

class Raster2eps
  def initialize( file )
    fp = 
      case( file )
      when( IO     ) then file
      when( String ) then File.open( file, "rb" )
      else
        raise StandardError, "argument must be IO or filename."
      end

    # read a raster image
    @format = "undef"
    @data   = ""
    @plte   = ""
    @width  = -1
    @height = -1
    @depth  = -1
    @type   = -1

    buf = fp.read( 8 )
    if   ( buf[0..7] == "\x89PNG\r\n\x1a\n" )   # PNG
      @format = "PNG"
      
      @data = ""
      while( ! fp.eof? )
        size = fp.read( 4 ).unpack( "N" )[0]
        name = fp.read( 4 )
        bin  = fp.read( size )
        crc  = fp.read( 4 ).unpack( "N" )[0]
        case( name )
        when( "IHDR" )
          @width, @height, @depth, @type, = bin.unpack( "N2C5" )
        when( "PLTE" )
          @plte = bin
        when( "IDAT" )
          @data << bin
        end
      end
    elsif( buf[0..3] == "\xff\xd8\xff\xe0" )    # JPEG
      @format = "JPEG"
      @data = buf << fp.read
      
      n = @data.length
      i = 2
      while( i < n )
        mark = @data[i,2]
        size = @data[i+2,2].unpack( "n" )[0]
        if( mark == "\xff\xc0" )   # SOF0
          @height, @width, @type, = @data[i+2,size].unpack( "x2xnnc" )
          break
        end
        i += 2 + size
      end
      @depth = 8
    else
      @format = "unknown"
      raise StandardError, "Unknown format.  dump: #{buf[0..7].inspect}..."
    end

    fp.close if file.kind_of?( String )

    # EPS settings
    @trans  = [ 0, 0 ]
    @rotate = 0
    @scale  = [ @width, @height ]
    @interpolate = true
  end

  def inspect
    "#<#{self.class}:0x#{sprintf( "%014x", self.object_id << 1 )} " << 
      ( self.instance_variables - [ :@data, :@plte ] ).collect!{ |symbol|
      "#{symbol}=#{self.instance_variable_get(symbol).inspect}"
    }.join( ", " ) << ">"
  end

  def palette
    @plte.dup
  end

  def palette=( plte )
    if( plte.length == @plte.length )
      @plte = plte.dup
    else
      STDERR.puts( "Palette size must be same (#{@plte.length} bytes)." )
      nil
    end
  end

  def convert
    clrspc, decode = 
      case( @format )   # colorspace, decode array
      when( "PNG"  )
        case( @type )
        when( 0 ) then [ "/DeviceGray", [ 0, 1 ]     ]   # grayscale
        when( 2 ) then [ "/DeviceRGB" , [ 0, 1 ] * 3 ]   # true color
        when( 3 )   # indexed color
          [ "[ /Indexed /DeviceRGB #{@plte.length/3-1} Palette ]", [ 0, 2**@depth-1 ] ]
        when( 4 )   # grayscale + alpha
          STDERR.puts( "Alpha channels are not supported." )
          return nil
        when( 6 )   # true color + alpha
          STDERR.puts( "Alpha channels are not supported." )
          return nil
        else
          STDERR.puts( "Invalid color type: #{@type}" )
          return nil
        end
      when( "JPEG" )
        case( @type )
        when( 1 ) then [ "/DeviceGray", [ 0, 1 ]     ]   # grayscale
        when( 3 ) then [ "/DeviceRGB" , [ 0, 1 ] * 3 ]   # YCbCr / YIQ
        when( 4 ) then [ "/DeviceCMYK", [ 0, 1 ] * 4 ]   # CMYK
        else
          STDERR.puts( "Invalid color type: #{@type}" )
          return nil
        end
      end
    filter = 
      case( @format )
      when( "PNG"  ) then "
<< /Predictor 10
   /Columns #{@width}
   /Colors #{decode.length/2}
   /BitsPerComponent #{@depth}
>> /FlateDecode filter"
      when( "JPEG" ) then "/DCTDecode filter"
      end

    # coordinate transformation
    buf = <<-EOS
%%BeginObject: #{@format}
gsave
#{@trans[0]} #{@trans[1]} translate
#{@rotate} rotate
#{@scale[0]} #{@scale[1]} scale
    EOS

    # palette
    if( @plte.length > 0 )
      i = 0
      hex = @plte.unpack( "H6" * ( @plte.length / 3 ) ).inject{ |str,rgb|
        str << ( ( i += 1 ) % 8 == 0 ? "\n" : " " ) << rgb
      }
      buf << <<-EOS
/Palette currentfile #{@plte.length} string readhexstring 
#{hex}
pop def
      EOS
    end

    # colorspace
    buf << "#{clrspc} setcolorspace\n"

    # image data
    buf << <<-EOS
/Data currentfile /ASCII85Decode filter #{filter} def
<< /ImageType 1
   /Width #{@width}
   /Height #{@height}
   /ImageMatrix [ #{@width} 0 0 #{-@height} 0 #{@height} ]
   /DataSource Data
   /BitsPerComponent #{@depth}
   /Decode [ #{decode.join(" ")} ]
   /Interpolate #{@interpolate}
>> image
    EOS
    buf << Ascii85.encode( @data ).scan( /.{1,60}/ ).join( "\n" ) << "~>\n"

    # coordinate transformation (restore)
    buf << <<-EOS
grestore
%%EndObject
    EOS

    return buf
  end

  attr_reader   :format, :width, :height, :depth, :type
  attr_accessor :trans, :rotate, :scale, :interpolate
end



if( $0 == __FILE__ )
  if( ARGV.length < 1 )
    STDERR.puts( "usage:  ruby raster2eps.rb input.(jpg|png) output.eps" )
    exit
  end

  fname_in  = ARGV[0]
  fname_out = ARGV[1]

  image = Raster2eps.new( fname_in == "-" ? STDIN : fname_in )

  # show image information
  STDERR.puts( " format : #{image.format}" )
  STDERR.puts( " size   : #{image.width}x#{image.height} px" )
  exit if( ARGV.length < 2 )


  # make postscript
  obj = image.convert
  exit unless( obj )
  
  fp_out = ( fname_out == "-" ? STDOUT : File.open( fname_out, "w" ) )
  fp_out.puts( <<-EOS )
%!PS-Adobe-3.0 EPSF-3.0
%%Title: #{fname_in}
%%Creator: raster2eps.rb
%%CreationDate: #{Time.now}
%%BoundingBox: 0 0 #{image.width} #{image.height}
%%DocumentData: Clean7Bit
%%LanguageLevel: 3
%%EndComments
%%BeginProlog
%%EndProlog
%%Page: 1 1
  EOS
  fp_out.puts( obj )
  fp_out.puts( "%%EOF\n" )
  fp_out.close if( fp_out.kind_of?( File ) )
end
